# -*- coding: utf-8 -*-
from .base import JayaBase
from .clasic import JayaClasic
from pyJaya.population import Population
from pyJaya.solution import Solution
import numpy as np


class JayaSelfAdadtive(JayaBase):

    def __init__(self, listVars, functionToEvaluate, listConstraints=[]):
        super(JayaSelfAdadtive, self).__init__(
            len(listVars)*10, listVars, functionToEvaluate, listConstraints)

    def nextPopulation(self, population):
        numOldSolutions = population.size()
        numNewSolutions = int(round(numOldSolutions*(1 + np.random.rand()-0.5)))

        if numNewSolutions == numOldSolutions:
            return population
        else:
            newPopulation = Population(self.minimax)
            if numNewSolutions < numOldSolutions:
                if numNewSolutions < self.cantVars:
                    numNewSolutions = self.cantVars
                if self.minimax:
                    for solution in population.sorted()[-numNewSolutions:]:
                        newPopulation.solutions.append(solution)
                else:
                    for solution in population.sorted()[:numNewSolutions]:
                        newPopulation.solutions.append(solution)
            elif numNewSolutions > numOldSolutions:
                for solution in population.solutions:
                    newPopulation.solutions.append(solution)
                if self.minimax:
                    for solution in population.sorted()[numOldSolutions-numNewSolutions:]:
                        newPopulation.solutions.append(solution)
                else:
                    for solution in population.sorted()[:numNewSolutions-numOldSolutions]:
                        newPopulation.solutions.append(solution)
            return newPopulation

    def run(self, numIterations):
        for i in range(numIterations):
            if i > 0:
                self.population = self.nextPopulation(self.population)
            self.population = JayaClasic(
                self.population.size(), self.listVars,
                self.functionToEvaluate, self.listConstraints,
                self.population).run(1)

        return self.population



# -*- coding: utf-8 -*-
from .base import JayaBase
from .clasic import JayaClasic
from pyJaya.population import Population
from pyJaya.solution import Solution
import numpy as np


class JayaQuasiOppositional(JayaBase):

    def generateQuasiOpposite(self, population):
        newPopulation = Population(self.minimax)
        for solution in population.solutions:
            newSolution = Solution(
                self.listVars, self.functionToEvaluate, self.listConstraints)
            newsolution = []
            for solt_item, solt_value in enumerate(solution.solution):
                AL = self.listVars[solt_item].minor
                AU = self.listVars[solt_item].major
                a = self.listVars[solt_item].convert((AL + AU) / 2)
                b = self.listVars[solt_item].convert(AL + AU - solt_value)
                val = (max(a, b) - min(a, b)) * np.random.random_sample() + min(a, b)
                newsolution.append(self.listVars[solt_item].convert(val))
            newSolution.setSolution(newsolution)
            newPopulation.solutions.append(newSolution)
        return newPopulation

    def newPopulation(self):
        auxPopulation = Population(self.minimax)
        quasiOppositePopulation = self.generateQuasiOpposite(self.population)
        for s in self.population.solutions:
            auxPopulation.solutions.append(s)
        for s in quasiOppositePopulation.solutions:
            auxPopulation.solutions.append(s)
        newPopulation = Population(self.minimax)
        if self.minimax:
            for s in auxPopulation.sorted()[-self.numSolutions:]:
                newPopulation.solutions.append(s)
        else:
            for s in auxPopulation.sorted()[:self.numSolutions]:
                newPopulation.solutions.append(s)
        return newPopulation

    def run(self, numIterations):
        for i in range(numIterations):
            self.population = self.newPopulation()
            self.population = JayaClasic(
                self.population.size(), self.listVars,
                self.functionToEvaluate, self.listConstraints,
                self.population).run(1)

        return self.population



# -*- coding: utf-8 -*-
from pyJaya.variants.base import JayaBase
from pyJaya.population import Population
from pyJaya.solution import Solution
import numpy as np


class JayaClasic(JayaBase):

    def run(self, number_iterations):
        for i in range(number_iterations):
            result = self.population.getBestAndWorst()
            np.random.seed()
            r1 = np.random.rand(self.cantVars)
            r2 = np.random.rand(self.cantVars)
            for solution in self.population.solutions:
                solt = []
                for v_item, v_value in enumerate(solution.solution):
                    solt.append(self.listVars[v_item].convert(
                        (v_value + r1[v_item] * (result['best_solution'][v_item] - abs(v_value)) - r2[v_item] * (result['worst_solution'][v_item]-abs(v_value)))
                    ))
                auxSolution = Solution(
                    self.listVars, self.functionToEvaluate,
                    self.listConstraints)
                auxSolution.setSolution(np.array(solt))
                if self.minimax:
                    if (auxSolution.value > solution.value) and \
                            (auxSolution.constraintsOK(np.array(solt))):
                        solution.setSolution(auxSolution.solution)
                else:
                    if (auxSolution.value < solution.value) and \
                            (auxSolution.constraintsOK(np.array(solt))):
                        solution.setSolution(auxSolution.solution)

        return self.population



# -*- coding: utf-8 -*-
"""
Solution class
"""
import numpy as np


class Solution():

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return str(self.solution)

    def __init__(self, listVars, functionToEvaluate, listConstraints=[]):

        self.cantVars = len(listVars)
        self.listVars = listVars
        self.listConstraints = listConstraints
        self.functionToEvaluate = functionToEvaluate

        self.solution = np.zeros(self.cantVars)
        self.value = None

    def generate(self):
        while 1:
            solution = np.zeros(self.cantVars)
            for item in range(self.cantVars):
                solution[item] = self.listVars[item].get()
            if self.constraintsOK(solution):
                self.solution = solution
                break
        self.value = self.evaluate()
        # print("solution.generate:", self.solution)

    def constraintsOK(self, solution):
        for constraints in self.listConstraints:
            if not constraints(*[solution]):
                return False
        return True

    def evaluate(self):
        return self.functionToEvaluate(*[self.solution])

    def setSolution(self, solution):
        self.solution = solution
        self.value = self.evaluate()



# -*- coding: utf-8 -*-
from .base import JayaBase
from .clasic import JayaClasic
from pyJaya.population import Population
from pyJaya.solution import Solution
from multiprocessing import Process, Queue, current_process, freeze_support
from multiprocessing import Pool
import numpy as np


class JayaSAMPE(JayaBase):

    def sprint(self, population):
        # result = population.getBestAndWorst()
        # solutions = list()

        numSolutions = len(population.solutions)
        jaya_clasic = JayaClasic(
            numSolutions, self.listVars, self.functionToEvaluate,
            population=population)
        if self.minimax:
            jaya_clasic.toMaximize()

        population = jaya_clasic.run(1)
        return population

    @staticmethod
    def worker(sampe, population):
        return sampe.sprint(population)

    def generate(self, m):
        entrada = self.population.divideInToWithElitist(m)
        # print("     ### entradas (luego de la division con elitismo)")
        # [print(e.solutions) for e in entrada]
        # print("     ###")
        pool = Pool(processes=3)
        results = [
            pool.apply_async(
                JayaSAMPE.worker, args=(self, x)) for x in entrada]
        subPopulations = [p.get() for p in results]

        newPopulation = Population(self.minimax)
        newPopulation.merge(subPopulations)
        # new_best_value = newPopulation.getBestAndWorst()['best_value']
        # old_best_value = self.population.getBestAndWorst()['best_value']
        #
        # if self.minimax:
        #     if new_best_value > old_best_value:
        #         self.population = newPopulation
        # else:
        #     if new_best_value < old_best_value:
        #         self.population = newPopulation
        #         self.population = newPopulation
        # newPopulation.getBestAndWorst()
        self.population = newPopulation

    def run(self, number_iterations):
        result = self.population.getBestAndWorst()
        bestValue = result['best_value']
        m = 2
        for i in range(number_iterations):
            # print(i, self.population.solutions)
            if i == 0:
                # print("Generaring", m)
                self.generate(2)
            else:
                if self.minimax:
                    bV = self.population.getBestAndWorst()['best_value']
                    if bV > bestValue:
                        if m < self.numSolutions:
                            m += 1
                        bestValue = bV
                    elif m > 2:
                        m -= 1
                else:
                    bV = self.population.getBestAndWorst()['best_value']
                    if bV < bestValue:
                        if m <= self.numSolutions:
                            m += 1
                        bestValue = bV
                    elif m > 2:
                        m -= 1
                # print("Generaring", m)
                self.generate(m)
                # print("done", m)
        return self.population.getBestAndWorst()



# -*- coding: utf-8 -*-
import numpy as np


class NumberRange():

    def __init__(self, outside_first=0, outside_second=1):

        self.minor = min(outside_first, outside_second)
        self.major = max(outside_first, outside_second)
        self.validate()

    def validate(self):
        """Client must define it self"""
        raise NotImplementedError("Client must define it self")

    def get(self):
        """Client must define it self"""
        raise NotImplementedError("Client must define it self")

    def convert(self):
        """Client must define it self"""
        raise NotImplementedError("Client must define it self")


class IntRange(NumberRange):

    def validate(self):
        if not isinstance(self.minor, (int)) or not isinstance(self.major, (int)):
            raise NotImplementedError("The numbers should be integer")

    def get(self):
        np.random.seed()
        return int((self.major-self.minor) * np.random.rand() + self.minor)

    def convert(self, value):
        value = round(value)
        if value > self.major:
            return self.major
        elif value < self.minor:
            return self.minor
        else:
            return value


class FloatRange(NumberRange):

    def validate(self):
        if not isinstance(
                self.minor, (float)) or not isinstance(self.major, (float)):
            raise NotImplementedError("The numbers should be float")

    def get(self):
        np.random.seed()
        return (self.major-self.minor) * np.random.rand() + self.minor

    def convert(self, value):
        if value > self.major:
            return self.major
        elif value < self.minor:
            return self.minor
        else:
            return value


class BinaryRange(NumberRange):

    def validate(self):
        pass

    def get(self):
        np.random.seed()
        return np.random.randint(2)

    def convert(self, value):
        value = round(value)
        if value > 1:
            return 1.0
        elif value < 0.0:
            return 0.0
        else:
            return value



# -*- coding: utf-8 -*-
from .base import JayaBase
from pyJaya.population import Population
from pyJaya.solution import Solution
import numpy as np


class JayaSAMP(JayaBase):

    def sprint(self, population):
        result = population.getBestAndWorst()
        r1 = np.random.rand(self.cantVars)
        r2 = np.random.rand(self.cantVars)
        for solution in population.solutions:
            solt = []
            for v_item, v_value in enumerate(solution.solution):
                solt.append(self.listVars[v_item].convert(
                    (v_value+r1[v_item] * (result['best_solution'][v_item] - abs(v_value)) - r2[v_item] * (result['worst_solution'][v_item]-abs(v_value)))
                ))
            auxSolution = Solution(
                self.listVars, self.functionToEvaluate,
                self.listConstraints)
            auxSolution.setSolution(np.array(solt))
            if self.minimax:
                if (auxSolution.value > solution.value) and \
                        (auxSolution.constraintsOK(np.array(solt))):
                    solution = auxSolution
            else:
                if (auxSolution.value < solution.value) and \
                        (auxSolution.constraintsOK(np.array(solt))):
                    solution.setSolution(auxSolution.solution)
        return population

    def run(self, number_iterations):
        result = self.population.getBestAndWorst()
        bestValue = result['best_value']
        for i in range(number_iterations):
            if i == 0:
                m = 2
                subPopulations = self.population.divideInTo(m)
                for p in subPopulations:
                    p = self.sprint(p)
                newPopulation = Population(self.minimax)
                newPopulation.merge(subPopulations)
                if self.minimax:
                    if newPopulation.getBestAndWorst()['best_value'] > self.population.getBestAndWorst()['best_value']:
                        self.population = newPopulation
                else:
                    if newPopulation.getBestAndWorst()['best_value'] < self.population.getBestAndWorst()['best_value']:
                        self.population = newPopulation
            else:
                if self.minimax:
                    bV = self.population.getBestAndWorst()['best_value']
                    if bV > bestValue:
                        if m < self.numSolutions:
                            m += 1
                        bestValue = bV
                    elif m > 1:
                        m -= 1
                    subPopulations = self.population.divideInTo(m)
                    for p in subPopulations:
                        p = self.sprint(p)
                    newPopulation = Population(self.minimax)
                    newPopulation.merge(subPopulations)
                    if newPopulation.getBestAndWorst()['best_value'] > self.population.getBestAndWorst()['best_value']:
                        self.population = newPopulation
                else:
                    bV = self.population.getBestAndWorst()['best_value']
                    if bV < bestValue:
                        if m < self.numSolutions:
                            m += 1
                        bestValue = bV
                    elif m > 1:
                        m -= 1
                    subPopulations = self.population.divideInTo(m)
                    for p in subPopulations:
                        p = self.sprint(p)
                    newPopulation = Population(self.minimax)
                    newPopulation.merge(subPopulations)
                    if newPopulation.getBestAndWorst()['best_value'] < self.population.getBestAndWorst()['best_value']:
                        self.population = newPopulation
        return self.population.getBestAndWorst()



# -*- coding: utf-8 -*-
"""
Population class
"""
from pyJaya.solution import Solution
from pyJaya.consts import *
import numpy as np
import copy


class Population():
    """Population class

    Args:
        minimax (int): Find minimum or maximum (0: minimize or 1: maximize)
    """

    def __init__(self, minimax):
        """Population init"""
        self.solutions = []
        self.minimax = minimax

    def generate(
            self, numSolutions, listVars, functionToEvaluate,
            listConstraints):
        """[summary]

        Args:
            numSolutions ([type]): [description]
            listVars ([type]): [description]
            functionToEvaluate ([type]): [description]
            listConstraints ([type]): [description]
        """        
        for i in range(numSolutions):
            solution = Solution(
                listVars, functionToEvaluate, listConstraints)
            solution.generate()
            self.solutions.append(solution)

    def toMaximize(self):
        self.minimax = minimaxType['maximize']

    def sorted(self):
        return sorted(
            self.solutions, reverse=self.minimax,
            key=lambda solution: solution.value)

    def getBestAndWorst(self):
        solutionSorted = self.sorted()
        if self.minimax:
            return {
                'best_value': solutionSorted[-1].value,
                'worst_value': solutionSorted[0].value,
                'best_solution': solutionSorted[-1].solution,
                'worst_solution': solutionSorted[0].solution}
        else:
            return {
                'best_value': solutionSorted[0].value,
                'worst_value': solutionSorted[-1].value,
                'best_solution': solutionSorted[0].solution,
                'worst_solution': solutionSorted[-1].solution}

    def divideInTo(self, n=2):
        if n == 1:
            return [self]
        populations = []
        q, r = divmod(len(self.solutions), n)
        indices = [q*i + min(i, r) for i in range(n+1)]
        subLists = [self.sorted()[indices[i]:indices[i+1]] for i in range(n)]
        for miniList in subLists:
            newPopulation = Population(self.minimax)
            newPopulation.solutions = miniList
            populations.append(newPopulation)
        return populations

    def divideInToWithElitist(self, n=2):
        if n == 1:
            return [self]
        populations = []
        q, r = divmod(len(self.solutions), n)
        indices = [q*i + min(i, r) for i in range(n+1)]
        subLists = [self.sorted()[indices[i]:indices[i+1]] for i in range(n)]
        for miniList in subLists[:-1]:
            newPopulation = Population(self.minimax)
            newPopulation.solutions = miniList
            populations.append(newPopulation)
        newPopulation = Population(self.minimax)
        newPopulation.solutions = copy.deepcopy(populations[0].solutions)
        populations.append(newPopulation)
        return populations

    def merge(self, listPopulation):
        listSolutions = []
        for p in listPopulation:
            listSolutions += p.solutions
        self.solutions = sorted(
            listSolutions, reverse=self.minimax,
            key=lambda solution: solution.value)
    
    def size(self):
        return len(self.solutions)



# -*- coding: utf-8 -*-
from pyJaya.consts import *
import numpy as np
from pyJaya.population import Population
import copy


class JayaBase(object):

    def __init__(
            self, numSolutions, listVars, functionToEvaluate,
            listConstraints=[], population=None):
        super(JayaBase, self).__init__()
        self.functionToEvaluate = functionToEvaluate
        self.numSolutions = numSolutions
        self.listVars = listVars
        self.cantVars = len(listVars)
        self.minimax = minimaxType['minimize']
        self.listConstraints = listConstraints

        if population is None:
            self.population = self.generatePopulation()
        else:
            self.population = copy.deepcopy(population)

    def generatePopulation(self):
        population = Population(self.minimax)
        population.generate(
            self.numSolutions, self.listVars, self.functionToEvaluate,
            self.listConstraints)
        return population

    def addConstraint(self, constraintFuntion):
        self.listConstraints.append(constraintFuntion)

    def toMaximize(self):
        self.minimax = minimaxType['maximize']
        self.population.toMaximize()

    def getBestAndWorst(self):
        return self.population.getBestAndWorst()

    def run(self, number_iterations):
        """Client must define it self"""
        raise NotImplementedError("Client must define it self")
